/**
 * Hamburg University of Technology
 * Institute of Telematics
 *
 *  Created on: Apr 15, 2013
 *      Author: homac
 *
 *
 *
 * This file contains the wrapper functions for malloc.
 *
 * All management data required by the mmprof layer are
 * not considered in the memory consumption statistics.
 *
 * requires -ldl
 */


#define _GNU_SOURCE         /* See feature_test_macros(7) */


#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <stdint.h>
#include <assert.h>
#include <unistd.h>
#include <sys/syscall.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <gmp.h>

#include "config.h"
#include "spinlock.h"
#include "offheap.h"
#include "hiddenmm.h"
#include "origmm.h"
#include "log.h"
#include "mma.h"
#include "module.h"


int mma_state = 0;
__thread int module_ignore = 0;



int output_humanized = DEFAULT_OUTPUT_HUMANIZED;
const char* logfile = DEFAULT_OUTPUT_FILE;



int getenv_bool(const char* envname, int default_value) {
	const char* env = getenv(envname);
	if (env) {
		if (!strcmp(env, "true") || !strcmp(env, "yes") || !strcmp(env, "1")) {
			return 1;
		} else if (!strcmp(env, "false") || !strcmp(env, "no") || !strcmp(env, "0")) {
			return 0;
		} else {
			return -1;
		}
	}
	return default_value;
}


/**
 * Library initialization. This function is automatically called after
 * the library has been load the first time by the running process
 * (usually at process start).
 */
__attribute__((constructor (65535)))
void mma_init (void)
{

	if (!(mma_state & MMA_INITIALIZED)) {

		/*
		 * Determine replacements through environment variables
		 */
		int b = getenv_bool(ENV_OFFHEAP_SHARED, offheap_shared);
		if (b < 0) {
			fprintf(stderr, ENV_OFFHEAP_SHARED);
			fprintf(stderr, " must be in [0..1]. Resetting to default ");
			fprintf(stderr, MMA_STRVAL(DEFAULT_OFFHEAP_SHARED));
		} else {
			offheap_shared = b;
		}

		offheap_init();

		/*
		 * Tell gmp to use offheap until we have loaded the allocator.
		 */
		mp_set_memory_functions (offheap_malloc, offheap_realloc_mp, offheap_free_mp);


		const char* env = getenv(ENV_OUTPUT_FILE);
		if (env) logfile = env;

		int ret = log_init(logfile);
		if (ret != 0) {
			log_error("can't open log file '%s'",  logfile);
		}


		/**
		 * Name of the libc to be load.
		 */
		const char* library_name = DEFAULT_LIB_NAME;
		const char* library_malloc_name = MMA_STRVAL(DEFAULT_LIB_MALLOC_NAME);
		const char* library_calloc_name = MMA_STRVAL(DEFAULT_LIB_CALLOC_NAME);
		const char* library_realloc_name = MMA_STRVAL(DEFAULT_LIB_REALLOC_NAME);
		const char* library_free_name = MMA_STRVAL(DEFAULT_LIB_FREE_NAME);


		env = getenv(ENV_LIB_NAME);
		if (env != NULL) library_name = env;


		b = getenv_bool(ENV_OUTPUT_HUMANIZED, output_humanized);
		if (b < 0) {
			log_warning("%s must be in [0..1]. Resetting to default %d",  ENV_OUTPUT_HUMANIZED, DEFAULT_OUTPUT_HUMANIZED);
		} else {
			output_humanized = b;
		}

		log_info("mma version %s\n", MMA_VER);

		origmm_init(library_name, library_malloc_name, library_calloc_name, library_realloc_name, library_free_name);

		/*
		 * Tell gmp to use orignal malloc functions.
		 */
		mp_set_memory_functions (hidden_malloc, hidden_realloc_mp, hidden_free_mp);


        mma_state |= MMA_INITIALIZED;

        /*
         * Initialize statistics
         */
        module_init();

        b = getenv_bool(ENV_START_SUSPENDED, DEFAULT_START_SUSPENDED);
        if (b) mma_suspend();

	} else {
		log_error("multiple calls to mma_init");
	}
}


/**
 * Finalization function for the library.
 */
__attribute__((destructor (65535)))
void mma_fini (void)
{
	if (mma_state & MMA_INITIALIZED) {
		/* reset to uninitialized, to redirect malloc to pool again */
		mma_state = mma_state & (~MMA_INITIALIZED);
		module_fini();
		offheap_fini();
		log_fini();
	} else {
		log_error("multiple calls to mma_init");
	}
}

/**
 * activate logging
 */

int mma_resume(void) {
	// XXX: requires locking!
	if (!(mma_state & MMA_SUSPENDED)) {
		return -1;
	} else {
		module_resume();
		mma_state &= (~MMA_SUSPENDED);
	}
	return 0;
}

/**
 * deactivate logging
 */
int mma_suspend(void) {
	// XXX: requires locking!
	if (mma_state && MMA_SUSPENDED) {
		return -1;
	} else {
		module_suspend();
		mma_state &= (~MMA_SUSPENDED);
	}
	return 0;
}

