/**
 * Hamburg University of Technology
 * Institute of Telematics
 *
 *  Created on: Sep 17, 2013
 *      Author: homac
 */

/** \file mmoa.c - Offline analysis for mmprof traces.
 *
 */

/* tell khash and brothers that we use stdc malloc here. */
#define AC_USE_MALLOC

#include "log.h"
#include "stats.h"


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "profile.h"
#include "perfclk.h"

/**
 * Global statistics
 */

statistics_t local_stats;
statistics_t* stats = &local_stats;



__thread mpz_t mpz_tmp;


int output_humanized = 1;


/**
 * Function to make sure local data has been initialized
 */
void module_init (const mpz_t first_clock) {

	stats_init(first_clock);

	mpz_init(mpz_tmp);
}


void module_fini (const mpz_t last_clock) {


	mpz_init_set(mpz_tmp, last_clock);
	mpz_sub(mpz_tmp, mpz_tmp, local_stats.clock_start);

	if (output_humanized) {
		log_printf("Gathered memory statistics:\n");
		log_printf("duration [nsecs]: %Zd\n", mpz_tmp);
		log_printf("load_current [byte]: %ld\n", local_stats.load_current);
		log_printf("load_max [byte]: %ld\n", local_stats.load_max);
		log_printf("load_total [byte nsec]: %Zd\n", local_stats.load_total);
		mpz_div(mpz_tmp, local_stats.load_total, mpz_tmp);
		log_printf("load_avg [byte]: %Zd\n", mpz_tmp);
	} else {
		log_printf("%Zd %ld %ld %Zd ", mpz_tmp, local_stats.load_current, local_stats.load_max, local_stats.load_total);
		mpz_div(mpz_tmp, local_stats.load_total, mpz_tmp);
		log_printf(" %Zd\n", mpz_tmp);
	}

	log_flush();
}


int main(int argc, char** argv) {

	log_init("mma-analysed.log");

	const char* trace_file = "mma.log";

	FILE* f = fopen(trace_file, "r");
	if (f == NULL) {
		perror("fopen");
		exit(-1);
	}

	perf_clock_t current_clock;     /* event time stamp */
	mpz_t mpz_clock;
	int elems_read;
	int elems_expected = 0;
	pid_t tid;              /* thread id */
	int fid;                /* function id */

	void* ptr;              /* ptr parameter */
	void* newptr;           /* newptr parameter */
	size_t size;            /* size parameter */
	size_t num;             /* num parameter */
	size_t elem;            /* elem parameter */

	typedef unsigned int* ptr_t;
	int line = 0;
	while (EOF != (elems_read = fscanf(f, "%ld %ld %d %d", &current_clock.tv_sec, &current_clock.tv_nsec, &tid, &fid))) {
		line++;
		if (elems_read != 4) {
			elems_expected = 4;
			goto abort_read;
		}
		perf_clock_to_mpz (mpz_clock, &current_clock);
		if (line == 1) {
			module_init(mpz_clock);
		}
		switch(fid) {
		case FC_MALLOC:
			elems_expected = 2;
			if (elems_expected != (elems_read = fscanf(f, " %x %ld\n", (ptr_t)&ptr, &size))) goto abort_read;
			stats_malloc_nolock_clock(ptr, size, mpz_clock);
			break;
		case FC_CALLOC:
			elems_expected = 3;
			if (elems_expected != (elems_read = fscanf(f, " %x %ld %ld\n", (ptr_t)&ptr, &num, &elem))) goto abort_read;
			stats_malloc_nolock_clock(ptr, num*elem, mpz_clock);
			break;
		case FC_FREE:
			elems_expected = 1;
			if (elems_expected != (elems_read = fscanf(f, " %x\n", (ptr_t)&ptr))) goto abort_read;
			stats_free_nolock_clock(ptr, mpz_clock);
			break;
		case FC_REALLOC:
			elems_expected = 3;
			if (elems_expected != (elems_read = fscanf(f, " %x %x %ld\n", (ptr_t)&newptr, (ptr_t)&ptr, &size))) goto abort_read;
			stats_realloc_nolock_clock(newptr, ptr, size, mpz_clock);
			break;
		case FC_TMSTMP:
			elems_expected = 0;
			if (elems_expected != (elems_read = fscanf(f, "\n"))) goto abort_read;
			stats_tmstmp_nolock_clock(mpz_clock);
			break;
		}
		elems_expected = 0;
	}

abort_read:
	if (elems_read != EOF && elems_read < 0) {
		fprintf(stderr, "%s:%d:0: error: system error\n", trace_file, line);
		perror("fscanf");
		fclose(f);
		log_fini();
		exit(-2);
	} else if (elems_expected && elems_expected != elems_read) {
		fprintf(stderr, "%s:%d:0: error: can't interpret line.\n", trace_file, line);
		fclose(f);
		log_fini();
		exit(-3);
	}
	module_fini(mpz_clock);
	fclose(f);
	log_fini();
	return 0;
}
