/**
 * Hamburg University of Technology
 * Institute of Telematics
 *
 *  Created on: Apr 16, 2013
 *      Author: homac
 *
 *  This is a very very simple allocator which is fast but
 *  produces a lot of fragmentation. So, we don't recommend it
 *  for commercial software!
 *
 *  TODO: replace by proper allocator
 */



#include "offheap.h"

#include <sys/mman.h>
#include <string.h>
#include "log.h"




int offheap_shared = DEFAULT_OFFHEAP_SHARED;


offheap_t __offheap_in_data_segment__  = {OFFHEAP_SIZE,0,OFFHEAP_LOCK_INITIALIZER,{0}};
offheap_t* offheap = &__offheap_in_data_segment__;


#define PTR2BLK(__ptr__)  (((block_t*)(__ptr__))-1)
#define BLK2PTR(__blk__)  ((byte_t*)(__blk__+1))
#define BLKNEXT(__blk__)  ((block_t*)(BLK2PTR(__blk__)+__blk__->size))
#define PTRVALID(__ptr__) offheap_member(__ptr__)
#define BLKVALID(blk) PTRVALID(blk)
#define PTREND()   (offheap->mem + offheap->brk)

void offheap_init(void) {
	size_t size = OFFHEAP_SIZE;

	if (offheap_shared) {
		const int pagesize = sysconf(_SC_PAGE_SIZE);
		size = sizeof(offheap_t);
		int alignment = (pagesize - (size & (pagesize-1)));
		size += alignment;

		void* tmp = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, -1, 0);
		if (tmp ==  (void *) -1) {
			assert(!"mmap failed!");
			fprintf(stderr, "error initializing offheap in shared memory");
		} else {
			offheap = tmp;
			size -= (sizeof(offheap_t)-sizeof(__offheap_in_data_segment__.mem));
		}
	}
	offheap->size = size;
	memset(&offheap->lock, 0, sizeof(offheap_lock_t));
	offheap->brk = 0;
}

void offheap_fini(void) {
	log_info("used offheap: %d byte", offheap->brk);
}



void* offheap_recycle(size_t size) {
	block_t* b;
	void* ptr;
	offheap_lock_acquire(&offheap->lock);
	for (b = (block_t*)offheap->mem;
			BLKVALID(b);
			b = BLKNEXT(b))
	{
		if (b->free && b->size >= size) {
			break;
		}
	}

	if (BLKVALID(b)) {
		int remaining = b->size - size;
		if (remaining > (sizeof(block_t)+1)) {
			/* split */
			b->size = size;
			block_t* b_next = BLKNEXT(b);
			b_next->size = remaining;
			b_next->free = 1;
		}
		ptr = BLK2PTR(b);
	} else ptr = NULL;
	offheap_lock_release(&offheap->lock);
	return ptr;
}


void* offheap_malloc (size_t size) {
	void* ptr;
	size_t bruttosize = size + sizeof(block_t);

#if 1
	ptr = offheap_recycle(size);
	if (ptr) return ptr;
#endif

	offheap_lock_acquire(&offheap->lock);
	if (offheap->brk + bruttosize >= offheap->size) {
		ptr = NULL;
	} else {
		block_t* blk = (block_t*)(offheap->mem + offheap->brk);

		blk->size = size;
		blk->free = 0;
		ptr = BLK2PTR(blk);

		offheap->brk += bruttosize;
	}
	offheap_lock_release(&offheap->lock);
	assert(ptr && "Out of memory");
	return ptr;
}


void __offheap_merge_free_blocks (void* ptr) {
	/* trying to merge free adjacent blocks */
	block_t* blk = PTR2BLK(ptr);
	block_t* next;
	for (next = BLKNEXT(blk); BLKVALID(next) && next->free; next = BLKNEXT(next)) {
		blk->size += next->size;
	}
}


void offheap_free(void* ptr) {
	if (!ptr) {
		/* ignore */
		return;
	}

	if (PTRVALID(ptr)) {
		offheap_lock_acquire(&offheap->lock);

		PTR2BLK(ptr)->free = 1;

		__offheap_merge_free_blocks(ptr);

		/* try to decrease offheap break */
		byte_t* next = (byte_t*)BLKNEXT(PTR2BLK(ptr));
		assert(next<=PTREND());
		if (next == PTREND()) {
			offheap->brk -= (PTR2BLK(ptr)->size + sizeof(block_t));
		}

		offheap_lock_release(&offheap->lock);
	}
}


void* offheap_calloc(size_t num, size_t elm) {
	size_t size = num*elm;
	void* ptr = offheap_malloc(size);
	memset(ptr,0,size);
	return ptr;
}


void* offheap_realloc(void* oldptr, size_t size) {
	void* ptr = NULL;

	if (oldptr) {
		if (PTRVALID(oldptr)) {
			offheap_lock_acquire(&offheap->lock);
			__offheap_merge_free_blocks(oldptr);
			offheap_lock_release(&offheap->lock);

			if (PTR2BLK(oldptr)->size >= size) {
				return oldptr;
			}
		}
		ptr = offheap_malloc(size);
		memcpy(ptr,oldptr,PTR2BLK(oldptr)->size);
		offheap_free(oldptr);
	} else {
		ptr = offheap_malloc(size);
	}
	return ptr;
}

