/**
 * Access to performance clock and proper arithmetics.
 *
 * requires -lrt
 */


#ifndef __PERFCLK_H__
#define __PERFCLK_H__



#include <stdint.h>
#include <sys/types.h>
#include <sys/time.h>
#include <time.h>
#include <assert.h>
#include <stdio.h>
#include "config.h"



/**
 * {
 *     long     tv_sec;
 *     long     tv_nsec;
 * };
 */
typedef struct timespec perf_clock_t;

#define PERF_CLOCK_INITIALIZER {0L,0L}



/**
 * Source of time for all measurements
 */
static inline
void perf_clock (perf_clock_t* trg) {
#if DEBUG
	int res = clock_gettime(CLOCK_MONOTONIC, trg);
	if (res != 0) {
		perror ("clock_gettime");
		assert(0);
	}
#else
	clock_gettime(CLOCK_MONOTONIC, trg);
#endif
}

static inline
int64_t* perf_clock_usec (int64_t* trg) {
	perf_clock_t tmp;
	perf_clock(&tmp);
	*trg = tmp.tv_sec*1000000LL + tmp.tv_nsec/1000LL;
	return trg;
}


static inline
int64_t perf_clock_to_nsec_long (perf_clock_t* trg) {
	return (int64_t)trg->tv_sec*1000000000LL + trg->tv_nsec;
}

static inline
perf_clock_t* perf_clock_nsec_to_clock(perf_clock_t* trg, int64_t nsec) {
	trg->tv_sec = nsec /1000000000L;
	trg->tv_nsec = nsec%1000000000L;
	return trg;
}

static inline
int64_t perf_clock_to_usec_long (perf_clock_t* tm) {
	return (int64_t)tm->tv_sec * 1000000LL  +  tm->tv_nsec/1000;
}


static inline
perf_clock_t* perf_clock_usec_to_clock(perf_clock_t* trg, int64_t usec) {
	trg->tv_sec =  usec/1000000L;
	trg->tv_nsec = usec%1000000L;
	return trg;
}


static inline
uint64_t perf_clock_granularity() {
	perf_clock_t trg;
	int res = clock_getres(CLOCK_MONOTONIC, &trg);
	if (res != 0) {
		perror ("clock_getres");
		assert(0);
	}
	return perf_clock_to_nsec_long(&trg);
}



static inline
int perf_clock_eq(perf_clock_t* t1, perf_clock_t* t2) {
	return perf_clock_to_nsec_long(t1) == perf_clock_to_nsec_long(t2);
}

static inline
int perf_clock_ge(perf_clock_t* t1, perf_clock_t* t2) {
	return perf_clock_to_nsec_long(t1) >= perf_clock_to_nsec_long(t2);
}

static inline
int perf_clock_gt(perf_clock_t* t1, perf_clock_t* t2) {
	return perf_clock_to_nsec_long(t1) > perf_clock_to_nsec_long(t2);
}

static inline
int perf_clock_le(perf_clock_t* t1, perf_clock_t* t2) {
	return perf_clock_to_nsec_long(t1) <= perf_clock_to_nsec_long(t2);
}

static inline
int perf_clock_lt(perf_clock_t* t1, perf_clock_t* t2) {
	return perf_clock_to_nsec_long(t1) < perf_clock_to_nsec_long(t2);
}


static inline
void perf_clock_sub(perf_clock_t* trg, perf_clock_t* sub) {
	trg->tv_sec -= sub->tv_sec;
	if (sub->tv_nsec > trg->tv_nsec) {
		trg->tv_sec -= 1;
		trg->tv_nsec = trg->tv_nsec+1000000000L- sub->tv_nsec;
	} else {
		trg->tv_nsec -= sub->tv_nsec;
	}
	ASSERT(trg->tv_sec >= 0);
}

static inline
void perf_clock_add(perf_clock_t* trg, perf_clock_t* add) {
	const int MAX_NANOS = 1000000000L;
	trg->tv_nsec += add->tv_nsec;
	if (trg->tv_nsec >= MAX_NANOS) {
		/* overflow */
		trg->tv_nsec = trg->tv_nsec % MAX_NANOS;
		trg->tv_sec++;
	}
	trg->tv_sec += add->tv_sec;
	ASSERT(trg->tv_sec >= 0);
}

/**
 * Writes a formatted string in the given buffer.
 * Buffer size should be more than
 */
static inline
const char* perf_clock_usec_str(uint64_t perf_clock_usec_int64, char* buf, size_t size) {
	uint64_t tm = perf_clock_usec_int64;
	snprintf(buf, size, "%02lu:%02lu.%06lu", tm/60000000, (tm/1000000)%60, (tm%60000000)%1000000);
	return buf;
}

/**
 * Writes a formatted string in the given buffer.
 * Buffer size should be more than
 */
static inline
const char* perf_clock_nsec_str(uint64_t perf_clock_nsec_int64, char* buf, size_t size) {
	uint64_t tm = perf_clock_nsec_int64;
	snprintf(buf, size, "%02lu:%02lu.%09lu", tm/60000000000L, (tm/1000000000L)%60, (tm%60000000000L)%1000000000L);
	return buf;
}


/**
 * Writes a formatted string in the given buffer.
 * Buffer size should be more than
 */
static inline
const char* perf_clock_str(perf_clock_t* clk, char* buf, size_t size) {
	snprintf(buf, size, "%02lu:%02lu.%09lu", clk->tv_sec/60, (clk->tv_sec%60), (clk->tv_nsec%1000000000L));
	return buf;
}




#endif /* __PERFCLK_H__ */
