/**
 * Hamburg University of Technology
 * Institute of Telematics
 *
 *  Created on: Apr 16, 2013
 *      Author: homac
 *
 * Very simple implemenation of a spinlock.
 *
 * It does not provide recursive calls.
 */

#ifndef SRC_SPINLOCK_H_
#define SRC_SPINLOCK_H_

#include <assert.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/syscall.h>
#include <unistd.h>
#include <sched.h>


/**
 * This is a write memory fence (memory barrier) on hardware and software level.
 * It prevents the compiler from moving writes above this fence when optimizing the
 * order of memory accesses.
 * It even ensures on hardware level that all stores will be finished prior to the
 * next instruction.
 */
#define SFENCE() 	__asm volatile("sfence" ::: "memory")

#define CACHE_LINE_SIZE 64
/**
 * Simple non-recursive lock to be used with the lock_xxx functions.
 */
typedef volatile uint32_t lock_t  __attribute__ ((aligned (CACHE_LINE_SIZE)));

#define LOCK_INITIALIZER 0



static inline
void lock_acquire(volatile lock_t* lock) {
	while (!__sync_bool_compare_and_swap(lock, 0, 1)){
		sched_yield();
	}
}

static inline
void lock_release(volatile lock_t* lock) {
	assert(*lock ==1);
	SFENCE();
	*lock = 0;
}




/**
 * Recursive spin lock to be used with the lock_recursive_xxx functions.
 */
typedef struct {
	lock_t lock;
	uint32_t depth;
} lock_recursive_t;

#define LOCK_RECURSIVE_INITIALIZER {0,0}

static inline
void lock_recursive_acquire(volatile lock_recursive_t* lock) {
	pid_t tid = (pid_t) syscall (SYS_gettid);
	if (lock->lock != tid) {
		while (!__sync_bool_compare_and_swap(&lock->lock, 0, tid));
	}
	lock->depth++;
}

static inline
void lock_recursive_release(volatile lock_recursive_t* lock) {
#if DEBUG
	pid_t tid = (pid_t) syscall (SYS_gettid);
	assert(lock->lock == tid);
#endif
	--lock->depth;
	if (!lock->depth) {
		SFENCE();
		lock->lock = 0;
	}
}




#endif /* SRC_SPINLOCK_H_ */
