/**
 * Hamburg University of Technology
 * Institute of Telematics
 *
 *  Created on: Sep 17, 2013
 *      Author: homac
 */

#ifndef SRC_STATS_H_
#define SRC_STATS_H_

#include "gmp.h"
#include "config.h"
#include "khash.h"
#include "perfclk.h"
#include "spinlock.h"




KHASH_MAP_INIT_INT64(addrmap, size_t);
typedef khash_t(addrmap) addrmap_t;

typedef struct {

	lock_t 		lock;
	addrmap_t* 	addrmap;
	mpz_t 		clock_start;

	mpz_t 		clock_last_update   __attribute__ ((aligned (CACHE_LINE_SIZE)));
	size_t 		load_max;
	size_t 		load_current;
	mpz_t 	    load_total;

	int 		instances;          /* number of running processes using this shared data structure */
} statistics_t;

extern statistics_t* stats;

extern __thread mpz_t mpz_tmp;



static inline
void perf_clock_to_mpz (mpz_t mpz, perf_clock_t* clock) {
	mpz_init_set_ui(mpz,clock->tv_sec);
	mpz_mul_ui(mpz,mpz,1000000000UL);
	mpz_add_ui(mpz,mpz,clock->tv_nsec);
}

static inline
void perf_clock_mpz (mpz_t mpz) {
	perf_clock_t clock;
	perf_clock(&clock);
	mpz_init_set_ui(mpz,clock.tv_sec);
	mpz_mul_ui(mpz,mpz,1000000000UL);
	mpz_add_ui(mpz,mpz,clock.tv_nsec);
}


static inline
void stats_update_last_event_tmstmp(const mpz_t clock) {
	/* update last update time */
	mpz_set(stats->clock_last_update, clock);
}

static inline
void stats_update_load_total (const mpz_t clock) {
	/*
	 * throughput += current*(clock_current-clock_last_update)
	 */
	mpz_set(mpz_tmp, clock);
	mpz_sub(mpz_tmp, mpz_tmp, stats->clock_last_update);
	mpz_mul_ui(mpz_tmp, mpz_tmp, stats->load_current);
	mpz_add(stats->load_total, stats->load_total, mpz_tmp);

}

static inline
void stats_init (const mpz_t start_time) {
	stats->load_max = 0;
	stats->load_current = 0;
	stats->addrmap = kh_init(addrmap);
	kh_resize(addrmap, stats->addrmap, STATS_ADDRMAP_INITIAL_SIZE);
	mpz_init_set_ui(stats->load_total, 0);

	mpz_init_set(stats->clock_start, start_time);
	mpz_init_set(stats->clock_last_update, start_time);

	stats->instances = 1;
}


static inline
void stats_tmstmp_nolock_clock (const mpz_t current_clock) {
	stats_update_load_total(current_clock);
	stats_update_last_event_tmstmp(current_clock);
}

static inline
void stats_tmstmp_nolock(void) {
	mpz_t clock_current;
	perf_clock_mpz(clock_current);
	stats_tmstmp_nolock_clock(clock_current);
}


static inline
void stats_add_memory(void* ptr, size_t size) {
	stats->load_current+=size;
	if (stats->load_current > stats->load_max) stats->load_max = stats->load_current;
	int ret;
	khiter_t k = kh_put(addrmap, stats->addrmap, (size_t)ptr, &ret);
//	if (!ret) {
//		log_error("can't store memory statistics. Out of memory?");
//		abort();
//	}
	kh_val(stats->addrmap, k) = size;
}

static inline
void stats_sub_memory(void* ptr) {
	khiter_t k = kh_get(addrmap, stats->addrmap, (size_t)ptr);
	if (k != kh_end(stats->addrmap)) {
		/* update */
		stats->load_current -= kh_val(stats->addrmap, k);
		kh_del(addrmap, stats->addrmap, k);
	}
}

static inline
void stats_malloc_nolock_clock (void* ptr, size_t size, const mpz_t current_clock) {
	if (!ptr) return;

	stats_update_load_total(current_clock);

	stats_add_memory(ptr, size);
	stats_update_last_event_tmstmp(current_clock);
}

static inline
void stats_malloc_nolock (void* ptr, size_t size) {
	mpz_t clock_current;
	perf_clock_mpz(clock_current);

	stats_malloc_nolock_clock (ptr, size, clock_current);
}



static inline
void stats_free_nolock_clock(void* ptr, const const mpz_t current_clock) {
	if (!ptr) return;
	stats_update_load_total(current_clock);
	stats_sub_memory(ptr);
	stats_update_last_event_tmstmp(current_clock);
}

static inline
void stats_free_nolock(void* ptr) {
	mpz_t clock_current;
	perf_clock_mpz(clock_current);
	stats_free_nolock_clock(ptr, clock_current);
}



static inline
void stats_realloc_nolock_clock(void* newptr, void* ptr, size_t size, const mpz_t current_clock) {
	if (!ptr) return;
	stats_update_load_total(current_clock);

	stats_sub_memory(ptr);
	stats_add_memory(newptr, size);
	stats_update_last_event_tmstmp(current_clock);
}

static inline
void stats_realloc_nolock(void* newptr, void* ptr, size_t size) {
	mpz_t clock_current;
	perf_clock_mpz(clock_current);
	stats_realloc_nolock_clock(newptr, ptr, size, clock_current);
}


#endif /* SRC_STATS_H_ */
